/*

diag.c - Diagnostic code

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <string.h>
#include <time.h>

#ifdef WIN32
  #define WIN32_LEAN_AND_MEAN
  #define BOOLEAN BOOLEANx
  #include <windows.h>
  #undef BOOLEAN
#elif defined(UNIX)
  #include <unistd.h>
#endif

#include "types.h"
#include "diag.h"

/*...vdiag\46\h:0:*/
/*...e*/

/*...svars:0:*/
unsigned int diag_methods = 0;
BOOLEAN diag_flags[DIAG_COUNT];

#define	RING_SIZE 0x10000
static int produce = 0, consume = 0;
static char *ring[RING_SIZE];
/*...e*/

/*...smethodvals:0:*/
typedef struct
	{
	const char *name;
	unsigned int value;
	} METHODVAL;

static METHODVAL methodvals[] =
	{
		{"console"	,DIAGM_CONSOLE},
		{"file"		,DIAGM_FILE},
		{"ring"		,DIAGM_RING},
	};
/*...e*/
/*...sdiag_method_of:0:*/
unsigned int diag_method_of(const char *s)
	{
	int i;
	for ( i = 0; i < sizeof(methodvals)/sizeof(methodvals[0]); i++ )
		if ( !strcmp(s, methodvals[i].name) )
			return methodvals[i].value;
	return 0;
	}
/*...e*/

/*...sflagvals:0:*/
typedef struct
	{
	const char *name;
	unsigned int value;
	} FLAGVAL;

static FLAGVAL flagvals[] =
	{
		{"win-unknown-key"	,DIAG_WIN_UNKNOWN_KEY},
		{"win"			,DIAG_WIN_UNKNOWN_KEY},
		{"mem-iobyte"		,DIAG_MEM_IOBYTE},
		{"mem-dump"		,DIAG_MEM_DUMP},
		{"mem"			,DIAG_MEM_IOBYTE},
		{"mem"			,DIAG_MEM_DUMP},
		{"vid-status"		,DIAG_VID_STATUS},
		{"vid-registers"	,DIAG_VID_REGISTERS},
		{"vid-address"		,DIAG_VID_ADDRESS},
		{"vid-data"		,DIAG_VID_DATA},
		{"vid-refresh"		,DIAG_VID_REFRESH},
		{"vid-markers"		,DIAG_VID_MARKERS},
		{"vid"			,DIAG_VID_STATUS},
		{"vid"			,DIAG_VID_REGISTERS},
		{"vid"			,DIAG_VID_ADDRESS},
		{"vid"			,DIAG_VID_DATA},
		{"vid"			,DIAG_VID_REFRESH},
		{"vid"			,DIAG_VID_MARKERS},
		{"kbd-win-key"		,DIAG_KBD_WIN_KEY},
		{"kbd-sense"		,DIAG_KBD_SENSE},
		{"kbd-remap"		,DIAG_KBD_REMAP},
		{"kbd"			,DIAG_KBD_WIN_KEY},
		{"kbd"			,DIAG_KBD_SENSE},
		{"kbd"			,DIAG_KBD_REMAP},
		{"snd-registers"	,DIAG_SND_REGISTERS},
		{"snd-latency"		,DIAG_SND_LATENCY},
		{"snd"			,DIAG_SND_REGISTERS},
		{"snd"			,DIAG_SND_LATENCY},
		{"ctc-registers"	,DIAG_CTC_REGISTERS},
		{"ctc-pending"		,DIAG_CTC_PENDING},
		{"ctc-interrupt"	,DIAG_CTC_INTERRUPT},
		{"ctc"			,DIAG_CTC_REGISTERS},
		{"ctc"			,DIAG_CTC_PENDING},
		{"ctc"			,DIAG_CTC_INTERRUPT},
		{"mon-hw"		,DIAG_MON_HW},
		{"mon-kbd-win-key"	,DIAG_MON_KBD_WIN_KEY},
		{"mon-kbd-map-th"	,DIAG_MON_KBD_MAP_TH},
		{"mon"			,DIAG_MON_HW},
		{"mon"			,DIAG_MON_KBD_WIN_KEY},
		{"mon"			,DIAG_MON_KBD_MAP_TH},
		{"sdxfdc-ports"		,DIAG_SDXFDC_PORT},
		{"sdxfdc"		,DIAG_SDXFDC_PORT},
		{"sdxfdc-hw"		,DIAG_SDXFDC_HW},
		{"sdxfdc"		,DIAG_SDXFDC_HW},
		{"sdxfdc-data"		,DIAG_SDXFDC_DATA},
		{"sdxfdc"		,DIAG_SDXFDC_DATA},
		{"sdxfdc-status"	,DIAG_SDXFDC_STATUS},
		{"sdxfdc"		,DIAG_SDXFDC_STATUS},
		{"sidisc-file"		,DIAG_SIDISC_FILE},
		{"sidisc"		,DIAG_SIDISC_FILE},
		{"sidisc-address"	,DIAG_SIDISC_ADDRESS},
		{"sidisc"		,DIAG_SIDISC_ADDRESS},
		{"sidisc-data"		,DIAG_SIDISC_DATA},
		{"sidisc"		,DIAG_SIDISC_DATA},
		{"print"		,DIAG_PRINT},
 		{"cpm-driver"		,DIAG_CPM_DRIVER},
 		{"cpm-cbios"		,DIAG_CPM_CBIOS},
 		{"cpm-bdos-file"	,DIAG_CPM_BDOS_FILE},
 		{"cpm-bdos-other"	,DIAG_CPM_BDOS_OTHER},
 		{"cpm-bdos"		,DIAG_CPM_BDOS_FILE},
 		{"cpm-bdos"		,DIAG_CPM_BDOS_OTHER},
		{"cpm"			,DIAG_CPM_DRIVER},
		{"cpm"			,DIAG_CPM_CBIOS},
		{"cpm"			,DIAG_CPM_BDOS_FILE},
		{"cpm"			,DIAG_CPM_BDOS_OTHER},
		{"z80-instructions"	,DIAG_Z80_INSTRUCTIONS},
		{"z80-instructions-new"	,DIAG_Z80_INSTRUCTIONS_NEW}, /* undocumented */
		{"z80-interrupts"	,DIAG_Z80_INTERRUPTS},
		{"z80"			,DIAG_Z80_INSTRUCTIONS},
		{"z80"			,DIAG_Z80_INTERRUPTS},
		{"bad-port"		,DIAG_BAD_PORT},
		{"panel-hack"		,DIAG_PANEL_HACK},
		{"tape"			,DIAG_TAPE},
		{"speed"		,DIAG_SPEED},
		{"exit"			,DIAG_EXIT},
                {"hw-win"               ,DIAG_HW_WIN},
                {"hw-kbd"               ,DIAG_HW_KBD},
                {"hw"                   ,DIAG_HW_WIN},
                {"hw"                   ,DIAG_HW_KBD},
	};
/*...e*/
/*...sdiag_flag_of:0:*/
BOOLEAN diag_flag_of(const char *s)
	{
	int i;
	BOOLEAN found = FALSE;
	if ( !strcmp (s, "all") )
		{
		for ( i = 0; i < DIAG_COUNT; ++i )
			diag_flags[i] = TRUE;
		return TRUE;
		}
	for ( i = 0; i < sizeof(flagvals)/sizeof(flagvals[0]); i++ )
		{
		if ( !strcmp(s, flagvals[i].name) )
			{
			diag_flags[flagvals[i].value] = TRUE;
			found = TRUE;
			}
		}
	return found;
	}
/*...e*/

/*...sdiag_message:0:*/
void diag_message(unsigned int flag, const char *fmt, ...)
	{
	if ( diag_flags[flag] )
		{
		va_list	vars;
		char s[256+1];
		va_start(vars, fmt);
		vsprintf(s, fmt, vars);
		va_end(vars);
		if ( diag_methods & DIAGM_CONSOLE )
			{
			fputs(s, stdout);
			fputc('\n', stdout);
			fflush(stdout);
			}
		if ( diag_methods & DIAGM_FILE )
			{
			FILE *fp;
			if ( (fp = fopen("memu.log", "a+")) != NULL )
				{
				fputs(s, fp);
				fputc('\n', fp);
				fclose(fp);
				}
			}
		if ( diag_methods & DIAGM_RING )
			{
			char *s2;
			if ( ring[produce%RING_SIZE] != NULL )
				free(ring[produce%RING_SIZE]);
			s2 = malloc(strlen(s)+1);
			if ( s2 != NULL )
				{
				strcpy(s2, s);
				ring[produce%RING_SIZE] = s2;
				++produce;
				// if ( produce == RING_SIZE ) terminate ("Ring Full");
				if ( produce%RING_SIZE == consume%RING_SIZE )
					++consume;
				}
			else
				diag_methods &= ~DIAGM_RING;
			}
		}
	}
/*...e*/

/*...sdiag_init:0:*/
void diag_init(void)
	{
	int i;
	for ( i = 0; i < DIAG_COUNT; ++i )
	   diag_flags[i]  =  FALSE;
	diag_flags[DIAG_ALWAYS] =  TRUE;
	for ( i = 0; i < RING_SIZE; i++ )
		ring[i] = NULL;
	}
/*...e*/
/*...sdiag_term:0:*/
void diag_term(void)
	{
	if ( diag_methods & DIAGM_RING )
		{
		FILE *fp;
		if ( (fp = fopen("memu.ring", "w")) != NULL )
			{
			while ( consume < produce )
				{
				fputs(ring[(consume++)%RING_SIZE], fp);
				fputc('\n', fp);
				}
			fclose(fp);
			}
		}
	}
/*...e*/
