/*

memu.c - Memotech Emulator

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#if defined(AIX) || defined(LINUX) || defined(SUN) || defined(MACOSX)
#include <unistd.h>
#else
#include <io.h>
#endif
#include <time.h>

#include "Z80.h"
#include "types.h"
#include "diag.h"
#include "common.h"
#include "win.h"
#include "roms.h"
#include "mem.h"
#include "vid.h"
#include "kbd.h"
#include "snd.h"
#include "ctc.h"
#include "mon.h"
#include "sdxfdc.h"
#include "sid.h"
#include "printer.h"
#include "cpm.h"
#include "dis.h"
#include "ui.h"
#include "memu.h"

/*...vZ80\46\h:0:*/
/*...vtypes\46\h:0:*/
/*...vdiag\46\h:0:*/
/*...vcommon\46\h:0:*/
/*...vwin\46\h:0:*/
/*...vroms\46\h:0:*/
/*...vmem\46\h:0:*/
/*...vvid\46\h:0:*/
/*...vkbd\46\h:0:*/
/*...vsnd\46\h:0:*/
/*...vctc\46\h:0:*/
/*...vmon\46\h:0:*/
/*...vsdxfdc\46\h:0:*/
/*...vsid\46\h:0:*/
/*...vprinter\46\h:0:*/
/*...vcpm\46\h:0:*/
/*...vdis\46\h:0:*/
/*...vui\46\h:0:*/
/*...vmemu\46\h:0:*/
/*...e*/

/*...susage:0:*/
#ifdef ALT_USAGE
extern void ALT_USAGE (void);
#endif

#ifdef ALT_EXIT
extern void ALT_EXIT (int reason);
#endif

void usage(void)
	{
	fprintf(stderr, "usage: memu [flags]\n");
	fprintf(stderr, "flags: -iobyte iobyte       specify IOBYTE (initially 0x00)\n");
	fprintf(stderr, "       -addr addr           set the address (initially 0x0000)\n");
	fprintf(stderr, "       -mem file            load file at address\n");
	fprintf(stderr, "       -mem-blocks n        number of 16KB memory blocks (default 4)\n");
	fprintf(stderr, "       -mem-mtx500          equivelent to -mem-blocks 2\n");
	fprintf(stderr, "       -romX file           load ROM X from file\n");
	fprintf(stderr, "       -vid-win             emulate VDP and TV using a graphical window\n");
	fprintf(stderr, "       -vid-win-big,-v      make window twice as large (realistic size)\n");
	fprintf(stderr, "       -vid-win-hw-palette  use an alternate palette\n");
	fprintf(stderr, "       -snd-portaudio,-s    emulate sound chip using portaudio\n");
	fprintf(stderr, "       -snd-latency value   instruct portaudio to use a given latency\n");
	fprintf(stderr, "       -mon-win             emulate 80 column card using a graphical window\n");
	fprintf(stderr, "       -mon-win-big,-mw     make window twice as high (realistic aspect ratio)\n");
	fprintf(stderr, "       -mon-win-mono        green screen monochrome\n");
#ifndef NO_TH
	fprintf(stderr, "       -mon-th,-mt          emulate 80 column card using full screen text mode\n");
#endif
	fprintf(stderr, "       -mon-console,-mc     emulate 80 column card using console only\n");
	fprintf(stderr, "       -mon-console-nokey   keyboard status shows no keys pressed\n");
	fprintf(stderr, "       -mon-ignore-init     ignore writes to non-emulated registers\n");
	fprintf(stderr, "       -kbd-remap           remaps MTX keyboard (despite shift state)\n");
	fprintf(stderr, "       -kbd-country n       sets the country code switches to n (default 0)\n");
	fprintf(stderr, "       -sdx-tracks n        specify tracks of first drive (default 80)\n");
	fprintf(stderr, "       -sdx-mfloppy file    specify .mfloppy file in SDX first drive\n");
	fprintf(stderr, "       -sdx-tracks2 n       specify tracks of second drive (default 80)\n");
	fprintf(stderr, "       -sdx-mfloppy2 file   specify .mfloppy file in SDX second drive\n");
	fprintf(stderr, "       -sidisc-file n file  specify Silicon Disc content for a drive\n");
	fprintf(stderr, "       -prn-file file       specify file to receive printer output\n");
	fprintf(stderr, "       -tape-dir path       .mtx files are in this directory\n");
	fprintf(stderr, "       -tape-overwrite      SAVE can overwrite an existing file\n");
	fprintf(stderr, "       -tape-disable        don't patch INOUT to LOAD/SAVE/VERIFY .mtx files\n");
	fprintf(stderr, "       -cpm                 emulate CP/M BDOS\n");
	fprintf(stderr, "       -cpm-drive-a path    where CP/M BDOS finds A: files (default: .)\n");
	fprintf(stderr, "       -cpm-invert-case     invert between CP/M and host filenames\n");
	fprintf(stderr, "       -cpm-tail tail       construct CP/M command tail\n");
	fprintf(stderr, "       -cpm-open-hack       don't insist on EX,S1,S2,RC being 0\n");
	fprintf(stderr, "       -sdx                 SDX support\n");
	fprintf(stderr, "       -speed hz            set CPU speed (default is 4000000, ie: 4MHz)\n");
	fprintf(stderr, "       -fast                don't limit speed, run as fast as possible\n");
#ifdef ALT_USAGE
	ALT_USAGE ();
#endif
	fprintf(stderr, "       file.com tail ...    -cpm -iobyte 0x80 -addr 0x0100 -mem file.com\n");
	fprintf(stderr, "       file.run             -iobyte 0x00 -addr 0xAAAA (from header)\n");
	fprintf(stderr, "       file.mtx             subsequent LOAD/SAVE/VERIFY \"\" will use this file\n");
#ifdef ALT_EXIT
	ALT_EXIT(2);
#else
	exit(2);
#endif
	}
/*...e*/
/*...sread_file:0:*/
static int read_file(const char *fn, byte *buf, int buflen)
	{
	FILE *fp;
	long length;
	if ( (fp = fopen(fn, "rb")) == NULL )
		fatal("can't open %s", fn);
	fseek(fp, 0, SEEK_END);
	length = (long) ftell(fp);
	if ( length > (long) buflen )
		{
		fclose(fp);
		fatal("%s is %ld bytes and can only be %d bytes at most", fn, length, buflen);
		}
	fseek(fp, 0, SEEK_SET);
	fread(buf, 1, (size_t) length, fp);
	fclose(fp);
	return (int) length;
	}
/*...e*/
/*...sget_word:0:*/
static word get_word(byte *p)
	{
	return (word) p[0] + (((word)p[1])<<8);
	}
/*...e*/

/*...svars:0:*/
CFG cfg;

static Z80 z80;
static BOOLEAN moderate_speed = TRUE;
static BOOLEAN panel_hack = FALSE;

static byte *run_buf = NULL;
static word run_hdr_base;
static word run_hdr_length;

static byte tape_buf[0xfff0];
static word tape_len;

static byte last_trace = TRUE;
static byte no_trace[0x10000>>3] = { 0 };

static BOOLEAN loadmtx_hack = FALSE;
/*...e*/

/*...smtx_tape:0:*/
/* Logic for implementing the MTX LOAD command
   is taken from MacTX by Per Persson.
   Fixed to use CALCST from Paul Daniels.
   Logic for SAVE and VERIFY added in a similar fashion. */

/*...sget_tape_name:0:*/
static const char *get_tape_name(word addr, const char *fn_prefix, char *fn_buf)
	{
	int i;
	char *p = fn_buf;
	if ( fn_prefix != NULL )
		{
		strcpy(p, fn_prefix);
		strcat(p, "/");
		p += strlen(fn_buf);
		}
	for ( i = 0; i < 15; i++ )
		p[i] = mem_read_byte(addr+i);
	while ( i > 0 && p[i-1] == ' ' )
		--i;
	if ( i > 0 )
		strcpy(p+i, ".mtx");
	else if ( cfg.tape_fn != NULL )
		strcpy(p, cfg.tape_fn);
	else
		strcpy(p, "default.mtx");
	return fn_buf;
	}
/*...e*/

#define	MAX_TAPE_PREFIX 300
static char tape_name_fn_buf[MAX_TAPE_PREFIX+15+1+3+1];
static const char *tape_name_fn;
static BOOLEAN tape_overwrite = FALSE;

static void mtx_tape(Z80 *r)
	{
	word base   = r->HL.W;
	word length = r->DE.W;
	word calcst = mem_read_byte(0xfa81) + mem_read_byte(0xfa82) * 256;
	if ( mem_read_byte(0xfd68) == 0 )
		/* SAVE */
		{
		FILE *fp;
		if ( base == calcst && length == 20 )
			{
			tape_name_fn = get_tape_name(calcst+1, cfg.tape_name_prefix, tape_name_fn_buf);
			diag_message(DIAG_TAPE, "SAVE fn=%s", tape_name_fn);
			if ( !tape_overwrite )
				{
				if ( (fp = fopen(tape_name_fn, "rb")) != NULL )
					{
					fclose(fp);
					fatal("not allowed to overwrite file %s", tape_name_fn);
					}
				}
			remove(tape_name_fn);
			length -= 2; /* Work around a bug in the MTX ROM */
			}
		diag_message(DIAG_TAPE, "SAVE base=0x%04x length=0x%04x iobyte=0x%02x", base, length, mem_get_iobyte());
		if ( length > sizeof(tape_buf) )
			fatal("attempt to SAVE 0x%04x byte chunk to tape", length);
		mem_read_block(base, length, tape_buf);
		if ( (fp = fopen(tape_name_fn, "ab+")) == NULL )
			fatal("can't append to file %s", tape_name_fn);
		fwrite(tape_buf, 1, (size_t) length, fp);
		fclose(fp);
		}
	else if ( mem_read_byte(0xfd67) != 0 )
		/* VERIFY.
		   Normally, if verification fails, the MTX BASIC ROM
		   stops the tape, cleans up and does rst 0x28.
		   That rst instruction is at 0x0adb. */
		{
		if ( base == 0xc011 && length == 18 )
			{
			tape_name_fn = get_tape_name(0xc002, cfg.tape_name_prefix, tape_name_fn_buf);
			diag_message(DIAG_TAPE, "VERIFY fn=%s", tape_name_fn);
			tape_len = read_file(tape_name_fn, tape_buf, sizeof(tape_buf));
			}
		/* Then verify chunks as requested */
		diag_message(DIAG_TAPE, "VERIFY base=0x%04x length=0x%04x iobyte=0x%02x", base, length, mem_get_iobyte());
		if ( length > tape_len )
			r->PC.W = 0x0adb;
		else
			{
			int i;
			for ( i = 0; i < length; i++ )
				if ( tape_buf[i] != mem_read_byte(base+i) )
					{
					r->PC.W = 0x0adb;
					break;
					}
			tape_len -= length;
			memmove(tape_buf, tape_buf+length, tape_len);
			}
		}
	else
		/* LOAD */
		{
		if ( base == 0xc011 && length == 18 )
			/* Load header, so read whole file */
			{
			tape_name_fn = get_tape_name(0xc002, cfg.tape_name_prefix, tape_name_fn_buf);
			diag_message(DIAG_TAPE, "LOAD fn=%s", tape_name_fn);
			tape_len = read_file(tape_name_fn, tape_buf, sizeof(tape_buf));
			}
		/* Then return chunks as requested */
		diag_message(DIAG_TAPE, "LOAD base=0x%04x length=0x%04x iobyte=0x%02x", base, length, mem_get_iobyte());
		if ( length > tape_len )
			fatal("attempt to LOAD 0x%04x byte chunk from tape and only 0x%04x remaining", length, tape_len);
		mem_write_block(base, length, tape_buf);
		tape_len -= length;
		memmove(tape_buf, tape_buf+length, tape_len);
		}
	/* The real routine disables interrupts */
	ctc_out(0, 0xf0);
	ctc_out(0, 0x03);
	ctc_out(1, 0x03);
	ctc_out(2, 0x03);
	ctc_out(3, 0x03);
	vid_reti();
	/* Then re-enables the video interrupt */
	ctc_out(0, 0xa5);
	ctc_out(0, 0x7d);
	}
/*...e*/
/*...szx_tape:0:*/
/* Special hack for the "Z" ZX Spectrum Emulator.
   Z patches the ZX Spectrum tape routines to point to routines within itself
   which will load the same format of tape, but using MTX hardware instead.
   I have a patched version of Z which has ED FE patch instructions in
   these alternative routines, and as a result, we will be called here.
   Although we don't actually care here, the 17 byte header typically has
     a type byte
     10 bytes of name
     a word holding the length of the block
     4 other bytes
*/

#define	CASSETTE_FILE "cassette.zxtape"

static byte zx_buf[0xc000];
static int zx_buf_len = 0;

static void zx_tape(Z80 *r)
	{
	word base   = r->IX.W;
	word length = r->DE.W;
	if ( length > sizeof(zx_buf) )
		fatal("Z tape request for 0x%04x bytes", length);
	if ( r->PC.W-2 == 0xe833 )
		{
		if ( r->AF.B.l & C_FLAG )
			/* LOAD */
			{
			if ( r->AF.B.h == 0 )
				/* Header */
				{
				diag_message(DIAG_TAPE, "Z LOAD %", CASSETTE_FILE);
				zx_buf_len = read_file(CASSETTE_FILE, zx_buf, sizeof(zx_buf));
				}
			diag_message(DIAG_TAPE, "Z LOAD base=0x%04x length=0x%04x", base, length);
			if ( length > zx_buf_len )
				fatal("attempt to Z LOAD 0x%04x byte chunk from tape and only 0x%04x remaining", length, zx_buf_len);
			mem_write_block(base, length, zx_buf);
			zx_buf_len -= length;
			memmove(zx_buf, zx_buf+length, zx_buf_len);
			}
		else
			/* VERIFY */
			{
			int i;
			if ( r->AF.B.h == 0 )
				/* Header.
				   This code path doesn't make a lot of sense,
				   as you have to find the program before
				   you verify it, so headers are always
				   loaded, not verified. */
				{
				diag_message(DIAG_TAPE, "Z VERIFY %s", CASSETTE_FILE);
				zx_buf_len = read_file(CASSETTE_FILE, zx_buf, sizeof(zx_buf));
				}
			diag_message(DIAG_TAPE, "Z VERIFY base=0x%04x length=0x%04x", base, length);
			if ( length > zx_buf_len )
				fatal("attempt to Z VERIFY 0x%04x byte chunk from tape and only 0x%04x remaining", length, zx_buf_len);
			r->AF.B.l |= (C_FLAG|Z_FLAG);
			for ( i = 0; i < length; i++ )
				if ( zx_buf[i] != mem_read_byte(base+i) )
					{
					r->AF.B.l &= ~(C_FLAG|Z_FLAG);
					break;
					}
			zx_buf_len -= length;
			memmove(zx_buf, zx_buf+length, zx_buf_len);
			}
		}
	else
		/* SAVE */
		{
		FILE *fp;
		if ( r->AF.B.h == 0 )
			{
			diag_message(DIAG_TAPE, "Z SAVE %s", CASSETTE_FILE);
			remove(CASSETTE_FILE);
			}
		diag_message(DIAG_TAPE, "Z SAVE base=0x%04x length=0x%04x", base, length);
		if ( (fp = fopen(CASSETTE_FILE, "ab+")) == NULL )
			fatal("can't append to file Z.output");
		mem_read_block(base, length, zx_buf);
		fwrite(zx_buf, 1, (size_t) length, fp);
		fclose(fp);
		}
	}
/*...e*/
/*...ssetup_sdx:0:*/
static void setup_sdx(void)
	{
	byte b = mem_get_iobyte();
	mem_set_iobyte(0x30);

	/* Copy the ROM into place */
	mem_write_block(0x2000, 0x2000, rom_sdx);

	/* During system initialisation, or after ROM 3, init_rom is called.
	   Ensure init_rom doesn't call fdsc_initlz */
	mem_write_byte(0x21a0, 0x00);
	mem_write_byte(0x21a1, 0x00);
	mem_write_byte(0x21a2, 0x00);

	/* When running a USER command, user_command is called.
	   If BDOS isn't initialised, init_bdos is called.
	   init_bdos calls init_bdos_initlz.
	   Ensure init_bdos_initlz doesn't call fdsc_initlz. */
	mem_write_byte(0x20bd, 0x00);
	mem_write_byte(0x20be, 0x00);
	mem_write_byte(0x20bf, 0x00);
	/*  Ensure init_bdos doesn't call fdsc_rwgo. */
	mem_write_byte(0x203d, 0x00);
	mem_write_byte(0x203e, 0x00);
	mem_write_byte(0x203f, 0x00);
	/* It then calls init_bdos_initlz.
	/* init_bdos It then reads 35 blocks from track 0 to 0xd700.
	   Ensure init_bdos doesn't call
	     fdsc_blkrd
	     init_bdos_loaded
	     init_bdos_patch
	     init_bdos_patch2
	   and replace it with code to place our magic instruction. */
	mem_write_byte(0x2048, 0x21); /* LD HL,#D706 */
	mem_write_byte(0x2049, 0x06);
	mem_write_byte(0x204a, 0xd7);
	mem_write_byte(0x204b, 0x36); /* LD (HL),#ED */
	mem_write_byte(0x204c, 0xed);
	mem_write_byte(0x204d, 0x23); /* INC HL */
	mem_write_byte(0x204e, 0x36); /* LD (HL),#FE */
	mem_write_byte(0x204f, 0xfe);
	mem_write_byte(0x2050, 0x23); /* INC HL */
	mem_write_byte(0x2051, 0x36); /* LD (HL),#C9 */
	mem_write_byte(0x2052, 0xc9);
	mem_write_byte(0x2053, 0x00); /* NOP */
	
	/* Hook in our CP/M BDOS support */
	cpm_init_sdx();

	/* Ensure that USER SYSCOPY is safe */
	mem_write_byte(0x321e, 0xef); /* RST #28 */
	mem_write_byte(0x321f, 0x01); /* "Mistake" */

	/* Ensure that USER FORMAT is safe */
	mem_write_byte(0x33d8, 0xef); /* RST #28 */
	mem_write_byte(0x33d9, 0x01); /* "Mistake" */

	/* Ensure that USER STAT is safe */
	mem_write_byte(0x3983, 0xef); /* RST #28 */
	mem_write_byte(0x3984, 0x01); /* "Mistake" */

	mem_set_iobyte(b);
	}
/*...e*/

/*...sPatchZ80:0:*/
void PatchZ80(Z80 *r)
	{
	if ( mem_get_iobyte() & 0x80 )
		{
		if ( cpm_patch(r) )
			;
		else if ( loadmtx_hack && cpm_patch_sdx(r) )
			;
		else if ( r->PC.W-2 == 0xe833 || r->PC.W-2 == 0xe920 )
			zx_tape(r);
		else
			fatal("magic instruction unexpectedly encountered at 0x%04x", (word)(r->PC.W-2));
		}
	else
		{
		if ( cpm_patch_sdx(r) )
			;
		else if ( r->PC.W-2 == 0x3627 && run_buf != NULL )
			/* Hack used to load .RUN file into memory
			   when MTX BASIC has started itself up */
			{
			mem_set_iobyte(0x00);
				/* In a real USER RUN, the IOBYTE
				   reflects the SDX BASIC ROM slot */
			mem_write_byte(0x3627, 0xd3);
			mem_write_byte(0x3628, 0x05);
			mem_write_block(run_hdr_base, run_hdr_length, run_buf+4);
			free(run_buf);
			run_buf = NULL;
			r->IFF&=0xFE; /* Disable interrupts, otherwise the
			                 interrupts set up by MTX BASIC will
			                 continue to happen, and can interfere. */
			r->PC.W = run_hdr_base; /* Jump to the code we loaded */
			vid_reset();
			}
		else if ( r->PC.W-2 == 0x0AAE )
			mtx_tape(r);
		else
			fatal("magic instruction unexpectedly encountered at 0x%04x", (word)(r->PC.W-2));
		}
	}
/*...e*/
/*...sLoopZ80:0:*/
/* CTC usage :-

   Channel  Counter Input  Output
   -------  -------------  ------
   0        VPDINT         nothing
   1        4MHz/13 *      DART serial clock 0 *
   2        4MHz/13 *      DART serial clock 1 *
   3        cassette *     nothing

   The timer input is always 4MHz, to all channels.

   *=> not emulated

*/

static long long ms_last_win_handle_events = 0;
static long long ms_last_mon_refresh_blink = 0;
static long long ms_last_moderate = 0;
static long long ms_last_speed_check = 0;
static long clk_speed = 4000000;
static long clk_last_vid_refresh = 0;
static long clk_now = 0;
static long clk_skipped = 0;
static long clk_last_moderate = 0;
static long clk_last_speed_check = 0;
static BOOLEAN force_moderate = FALSE;

word LoopZ80(Z80 *r)
	{
	long long ms_now = get_millis();
	long clk_advanced;
	int periph_int;
	
	/* Ensure XWindows is kept happy */
	if ( ms_now - ms_last_win_handle_events > 10 )
		{
		win_handle_events();
		ms_last_win_handle_events = ms_now;
		}

	/* Ensure the cursor (and text) flashes on the monitor screen */
	if ( ms_now - ms_last_mon_refresh_blink > 50 )
		{
		mon_refresh_blink();
		ms_last_mon_refresh_blink = ms_now;
		}

	clk_advanced = r->IPeriod - clk_skipped;
	clk_skipped = 0;
	clk_now += clk_advanced;

	if ( moderate_speed )
		if ( force_moderate || ms_now - ms_last_moderate > 5 )
			{
			long clk = (long) (clk_now-clk_last_moderate);
			long ms_elapsed  = (long) (ms_now-ms_last_moderate);
			long ms_realtime = clk/(clk_speed/1000);
			if ( ms_elapsed < ms_realtime )
				delay_millis(ms_realtime-ms_elapsed);
			ms_last_moderate  = ms_now;
			clk_last_moderate = clk_now;
			force_moderate = FALSE;
			}

	if ( ms_now - ms_last_speed_check > 1000 )
		{
		if ( diag_flags[DIAG_SPEED] )
			{
			long clk = (long) (clk_now-clk_last_speed_check);
			long ms  = (long) (ms_now-ms_last_speed_check);
			diag_message(DIAG_SPEED, "Speed %4.2lfMHz",
				( ( clk * 1000.0 ) / ms ) / 1000000.0
				);
			}
		ms_last_speed_check = get_millis();
			/* We fetch this again as we know the
			   diag_message call is going to take
			   a non-trivial length of time. */
		clk_last_speed_check = clk_now;
		}

	if ( clk_now - clk_last_vid_refresh > clk_speed / 50 )
		{
		vid_refresh();
			/* This will cause the frame to be displayed,
			   and thus the F bit in the status register to be
			   set, and also, if IE, VPDINT to be asserted. */
		clk_last_vid_refresh = clk_now;
		force_moderate = TRUE;

		ui_refresh();
			/* Redraw UI windows 50 times a second as well. */
		}

	if ( vid_int_pending() )
		ctc_trigger(0);
			/* As the CTC is typically configured with channel 0
			   in counter mode, with a count of 1, this will likely
			   cause an interrupt to be asserted to the Z80. */

	/* Now feed the system clock into all CTC channels */
	ctc_advance(0, (int) clk_advanced);
	ctc_advance(1, (int) clk_advanced);
	ctc_advance(2, (int) clk_advanced);
	ctc_advance(3, (int) clk_advanced);

	if ( panel_hack )
		{
		diag_message(DIAG_PANEL_HACK, "panelHack FIRED pc=0x%04x ICount=%ld", r->PC.W, (long) r->ICount);
		periph_int = ( ctc_get_int_vector()|0x04 );
		panel_hack = FALSE;
		}
	else
		periph_int = ctc_int_pending();

	if ( periph_int != -1 )
		{
		static long clkp[4];
		if ( diag_flags[DIAG_Z80_INTERRUPTS] )
			{
			diag_message(DIAG_Z80_INTERRUPTS, "Z80 interrupt, periph=0x%02x at clk=%ld clk_diff=%ld",
				periph_int, (long) clk_now,
				(long)clk_now-clkp[(periph_int>>1)&3]
				);
			clkp[(periph_int>>1)&3] = (long) clk_now;
			}
		if ( periph_int == INT_NMI )
			/* This is to work around a bug in the Z80 CPU
			   implementation. The peripheral byte can
			   legitimately be 0x66, in which case it would be
			   combined with the I register to make the final
			   vector. However, Z80.c IntZ80() checks for NMI
			   ahead of combining with I.
			   The MTX Videowall code happens to trigger this
			   bug by accident/fluke. */
			{
			diag_message(DIAG_Z80_INTERRUPTS, "Z80 interrupt, Z80.c IntZ80() bug workaround, suppressing interrupt");
			periph_int = INT_NONE;
			}
		return periph_int;
			/* Will be combined with I register and a word
			   fetched from that address, as per Z80 IM 2 */
		}
	else
		return INT_NONE;
	}
/*...e*/
/*...sDebugZ80:0:*/
/*...sDebugZ80Instruction:0:*/
/* Note: P and V flags are one and the same. */

static void DebugZ80Instruction(Z80 *r, const char *instruction)
	{
	diag_message(DIAG_Z80_INSTRUCTIONS, "a=%02x f=%c%c%c%c%c%c bc=%04x de=%04x hl=%04x ix=%04x iy=%04x sp=%04x pc=%04x  %s",
		r->AF.B.h,
		(r->AF.B.l & S_FLAG)!=0?'S':'.',
		(r->AF.B.l & Z_FLAG)!=0?'Z':'.',
		(r->AF.B.l & H_FLAG)!=0?'H':'.',
		(r->AF.B.l & P_FLAG)!=0?'P':'.',
		(r->AF.B.l & N_FLAG)!=0?'N':'.',
		(r->AF.B.l & C_FLAG)!=0?'C':'.',
		r->BC.W,
		r->DE.W,
		r->HL.W,
		r->IX.W,
		r->IY.W,
		r->SP.W,
		r->PC.W,
		instruction
		);
	}
/*...e*/

byte DebugZ80(Z80 *r)
	{
	if ( diag_flags[DIAG_Z80_INSTRUCTIONS] )
		{
		word pc = r->PC.W;
		if ( ( no_trace[pc>>3]&(0x01<<(pc&7)) ) == 0 )
			{
			char buf[500+1];
			if ( diag_flags[DIAG_Z80_INSTRUCTIONS_NEW] )
				no_trace[pc>>3] |= (0x01<<(pc&7));
			dis_instruction(&pc, buf);
			DebugZ80Instruction(r, buf);
			last_trace = TRUE;
			}
		else if ( last_trace )
			{
			DebugZ80Instruction(r, "...");
			last_trace = FALSE;
			}
		}
	return 1;
	}
/*...e*/
/*...sRetiZ80:0:*/
void RetiZ80(Z80 *R)
	{
	vid_reti();
	}
/*...e*/
/*...sOutZ80:0:*/
/*...sOutZ80_bad:0:*/
static void OutZ80_bad(const char *hardware, word port, byte value, BOOLEAN stop)
	{
	diag_message(DIAG_BAD_PORT, "no emulation of %s, out 0x%04x,0x%02x", hardware, port, value);
	if ( stop && !diag_flags[DIAG_BAD_PORT] )
		fatal("no emulation of %s, so stopping emulation", hardware);
	}
/*...e*/

void OutZ80(word port, byte value)
	{
	switch ( port )
		{
		case 0:
			mem_out0(value);
			break;
		case 1:
			vid_out1(value);
			break;
		case 2:
			vid_out2(value);
			break;
		case 3:
			OutZ80_bad("cassette tape", port, value, FALSE);
			break;
		case 4:
			print_out4(value);
			break;
		case 5:
			kbd_out5(value);
			break;
		case 6:
			snd_out6(value);
			break;
		case 7:
			OutZ80_bad("PIO", port, value, FALSE);
			break;
		case 8:
		case 9:
		case 10:
		case 11:
			ctc_out(port&0x03, value);
			if ( (port&0x03) == 2 )
				/* Its a write to channel 2 */
				if ( ( z80.PC.W == 0x218c && (mem_get_iobyte()&0xf0) == 0x10 ) || /* MTX BASIC PANEL */
				     ( z80.PC.W == 0xa4fa && (mem_get_iobyte()&0x80) == 0x80 ) )  /* VDEB.COM */
/*...sPANEL hack:40:*/
/* Allow PANEL single step to work.
   It very cleverly programs channel 2 in timer mode with prescaler 16
   and value 13, which is 208 clks, which means that the instant the
   single instruction is executed, an interrupt will be pending.
   This means they can single step ROM, as they don't have to patch
   the following instruction.

   Normally, in MEMU, interrupts aren't checked for after every instruction.
   So actually, we'd "single step" several instructions at once.
   To avoid this, we reduce the ICount to just enough to start executing
   the single instruction, then call LoopZ80 immediately afterwards.

   The normal CTC operation in response to the previous register write
   will not have caused a reload of prescaler or counter.
   To be sure of consistent behaviour, we must cause the CTC to reload
   to exactly what we've asked for. */

{
ctc_reload(2); /* Get consistent behaviour */
clk_skipped = z80.ICount; /* Be sure not to forget to account for this time */
z80.ICount = 13*16+3; /* Took some fiddling to get this right */
panel_hack = TRUE;
diag_message(DIAG_PANEL_HACK, "panelHack ARMED pc=0x%04x", z80.PC.W);
}
/*...e*/
			break;
		case 12:
		case 13:
		case 14:
		case 15:
			OutZ80_bad("DART", port, value, FALSE);
			break;
		case 0x10:
		case 0x11:
		case 0x12:
		case 0x13:
		case 0x14:
			sdxfdc_out(port, value);
			break;
		case 0x1f:
			/* MTX BASIC LOAD/SAVE/VERIFY doesn't use this,
			   but certain cassette loader implementations do,
			   such as the one I had the source for (INOUT.MAC),
			   and the one used by Jet Set Willy. */
			switch ( value )
				{
				case 0xaa:
					diag_message(DIAG_TAPE, "Cassette remote control start");
					break;
				case 0x55:
					diag_message(DIAG_TAPE, "Cassette remote control stop");
					break;
				default:
					fatal("unexpected write of 0x%02x to cassette remote control port", value);
				}
			break;
		case 0x30:
			mon_out30(value);
			break;
		case 0x31:
			mon_out31(value);
			break;
		case 0x32:
			mon_out32(value);
			break;
		case 0x33:
			mon_out33(value);
			break;
		case 0x38:
			mon_out38(value);
			break;
		case 0x39:
			mon_out39(value);
			break;
		case 0x40:
		case 0x41:
		case 0x42:
		case 0x43:
		case 0x44:
		case 0x45:
		case 0x46:
		case 0x47:
			OutZ80_bad("SJM's FDX FDC", port, value, TRUE);
			break;
		case 0x50: /* Low address */
		case 0x51: /* High address */
		case 0x53: /* Data R/W */
		case 0x54:
		case 0x55:
		case 0x57:
		case 0x58:
		case 0x59:
		case 0x5b:
		case 0x5c:
		case 0x5d:
		case 0x5f:
			sid_out(port, value);
			break;
		case 0xc0:
		case 0xc1:
			OutZ80_bad("REMEMOTECH HEXs", port, value, TRUE);
		case 0xc4:
			OutZ80_bad("REMEMOTECH LEDGs", port, value, TRUE);
		case 0xd0:
		case 0xd1:
			OutZ80_bad("REMEMOTECH/REMEMOrizer page registers", port, value, TRUE);
		case 0xd4:
		case 0xd6:
			OutZ80_bad("REMEMOTECH/REMEMOrizer SD Card", port, value, TRUE);
		case 0xd9:
			OutZ80_bad("REMEMOrizer flags", port, value, TRUE);
		default:
			OutZ80_bad("unknown hardware", port, value, FALSE);
			break;
		}
	}
/*...e*/
/*...sInZ80:0:*/
/*...sInZ80_bad:0:*/
static byte InZ80_bad(const char *hardware, word port, BOOLEAN stop)
	{
	diag_message(DIAG_BAD_PORT, "no emulation of %s, in 0x%04x", hardware, port);
	if ( stop && !diag_flags[DIAG_BAD_PORT] )
		fatal("no emulation of %s, so stopping emulation", hardware);
	return 0xff;
	}
/*...e*/

byte InZ80(word port)
	{
	switch ( port )
		{
		case 0:
			return print_in0();
		case 1:
			return vid_in1();
		case 2:
			return vid_in2();
		case 3:
			return snd_in3();
		case 4:
			return print_in4();
		case 5:
			return kbd_in5();
		case 6:
			return kbd_in6();
		case 7:
			return InZ80_bad("PIO", port, FALSE);
		case 8:
		case 9:
		case 10:
		case 11:
			return ctc_in(port&0x03);
		case 12:
		case 13:
		case 14:
		case 15:
			return InZ80_bad("DART", port, TRUE);
		case 0x10:
		case 0x11:
		case 0x12:
		case 0x13:
		case 0x14:
			return sdxfdc_in(port);
		case 0x30:
			return mon_in30();
		case 0x32:
			return mon_in32();
		case 0x33:
			return mon_in33();
		case 0x38:
			return mon_in38();
		case 0x39:
			return mon_in39();
		case 0x40:
		case 0x41:
		case 0x42:
		case 0x43:
		case 0x44:
		case 0x45:
		case 0x46:
		case 0x47:
			return InZ80_bad("SJM's FDX FDC", port, TRUE);
		case 0x50:
		case 0x51:
		case 0x53:
		case 0x54:
		case 0x55:
		case 0x57:
		case 0x58:
		case 0x59:
		case 0x5b:
		case 0x5c:
		case 0x5d:
		case 0x5f:
			return sid_in(port);
		case 0xc0:
		case 0xc1:
			return InZ80_bad("REMEMOTECH HEXs", port, TRUE);
		case 0xc4:
			return InZ80_bad("REMEMOTECH LEDGs", port, TRUE);
		case 0xc5:
			return InZ80_bad("REMEMOTECH KEYs", port, TRUE);
		case 0xc7:
			return InZ80_bad("REMEMOTECH XKEYs", port, TRUE);
		case 0xd0:
		case 0xd1:
			return InZ80_bad("REMEMOTECH/REMEMOrizer page registers", port, TRUE);
		case 0xd4:
		case 0xd6:
		case 0xd7:
			return InZ80_bad("REMEMOTECH/REMEMOrizer SD Card", port, TRUE);
		case 0xd8:
			return InZ80_bad("REMEMOTECH/REMEMOrizer clock divider", port, TRUE);
		case 0xd9:
			return InZ80_bad("REMEMOrizer flags", port, TRUE);
		default:
			return InZ80_bad("unknown hardware", port, TRUE);
		}
	}
/*...e*/

/*...smemu_reset:0:*/
void memu_reset(void)
	{
	ctc_init();
	mem_set_iobyte (0x00);
	ResetZ80(&z80);
	}
/*...e*/

/*...smain:0:*/
/*...ssdxcfg:0:*/
static int sdxcfg(int tracks)
	{
	switch ( tracks )
		{
		case 40: return DRVS_DOUBLE_SIDED | DRVS_40TRACK | DRVS_1_DRIVE;
		case 80: return DRVS_DOUBLE_SIDED | DRVS_80TRACK | DRVS_1_DRIVE;
		}
	fatal("unsupported number of tracks %d", tracks);
	return -1; /* Not reached */
	}
/*...e*/

#ifdef ALT_OPTIONS
extern BOOLEAN ALT_OPTIONS (int *pargc, char ***pargv, int *pi);
#endif

#ifdef ALT_ENTRY
int ALT_ENTRY (int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
	{
	int i;
	unsigned addr = 0;
	memset (&cfg, 0, sizeof (cfg));
	cfg.vid_width_scale = 1;
	cfg.vid_height_scale = 1;
	cfg.mon_width_scale = 1;
	cfg.mon_height_scale = 1;
	cfg.iperiod = 1000;
	cfg.tracks_sdxfdc[0] = 80;
	cfg.tracks_sdxfdc[1] = 80;

	if ( argc == 1 )
		usage();

	diag_init();

	mem_init();
	dis_init();

	for ( i = 1; i < argc; i++ )
		{
		/* printf ("argv[%d] = \"%s\"\n", i, argv[i]); */
#ifdef ALT_OPTIONS
		if ( ALT_OPTIONS (&argc, &argv, &i) )
			;
		else
#endif
			if ( !strcmp(argv[i], "-iobyte") )
			{
			unsigned iobyte;
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%i", &iobyte);
			mem_set_iobyte((byte) iobyte);
			}
		else if ( !strcmp(argv[i], "-addr") )
			{
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%i", &addr);
			}
		else if ( !strcmp(argv[i], "-mem") )
			{
			byte buf[0x10000];
			int len;
			if ( ++i == argc )
				usage();
			len = read_file(argv[i], buf, sizeof(buf));
			mem_write_block((word) addr, (word) len, buf);
			}
		else if ( !strcmp(argv[i], "-mem-blocks") )
			{
			int nblocks;
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%i", &nblocks);
			mem_alloc(nblocks);
			}
		else if ( !strcmp(argv[i], "-mem-mtx500") )
			mem_mtx500();
		else if ( !strncmp(argv[i], "-rom", 4) )
			{
			byte *buf = emalloc(0x2000);
			int rom;
			byte iob;
			sscanf(argv[i]+4, "%i", &rom);
			if ( rom < 0 || rom > 7 )
				fatal("ROM must be between 0 and 7");
			if ( ++i == argc )
				usage();
			cfg.rom_fn[rom] = argv[i];
			if ( read_file(argv[i], buf, 0x2000) != 0x2000 )			
				fatal("ROM file must be 0x2000 bytes long");
			iob = mem_get_iobyte();
			mem_set_iobyte((byte)(rom<<4));
			mem_write_block(0x2000, 0x2000, buf);
			mem_set_iobyte(iob);
			free(buf);
			}
		else if ( !strcmp(argv[i], "-kbd-remap") )
			cfg.kbd_emu |= KBDEMU_REMAP;
		else if ( !strcmp(argv[i], "-kbd-country") )
			{
			int country;
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%i", &country);
			if ( ( country < 0 ) || ( country > 3 ) )
				usage ();
			cfg.kbd_emu |= ( country << 2 );
			}
		else if ( !strcmp(argv[i], "-vid-win") )
			cfg.vid_emu |= VIDEMU_WIN;
		else if ( !strcmp(argv[i], "-vid-win-big") ||
			  !strcmp(argv[i], "-v")           )
			{
			cfg.vid_emu |= VIDEMU_WIN;
			++cfg.vid_width_scale;
			++cfg.vid_height_scale;
			}
		else if ( !strcmp(argv[i], "-vid-win-hw-palette") )
			cfg.vid_emu |= (VIDEMU_WIN|VIDEMU_WIN_HW_PALETTE);
		else if ( !strcmp(argv[i], "-snd-portaudio") ||
			  !strcmp(argv[i], "-s")             )
			cfg.snd_emu |= SNDEMU_PORTAUDIO;
		else if ( !strcmp(argv[i], "-snd-latency") )
			{
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%lf", &cfg.latency);
			}
		else if ( !strcmp(argv[i], "-mon-win") )
			cfg.mon_emu |= MONEMU_WIN;
		else if ( !strcmp(argv[i], "-mon-win-big") ||
			  !strcmp(argv[i], "-mw")          )
			{
			cfg.mon_emu |= MONEMU_WIN;
			++cfg.mon_height_scale;
			cfg.mon_width_scale = cfg.mon_height_scale/2;
			}
		else if ( !strcmp(argv[i], "-mon-win-mono") )
			cfg.mon_emu |= MONEMU_WIN_MONO;
		else if ( !strcmp(argv[i], "-mon-th") ||
			  !strcmp(argv[i], "-mt")     )
#ifndef NO_TH
			cfg.mon_emu |= MONEMU_TH;
#else
			fatal("this version of MEMU compiled without TH support");		
#endif
		else if ( !strcmp(argv[i], "-mon-console") ||
			  !strcmp(argv[i], "-mc")          )
			cfg.mon_emu |= MONEMU_CONSOLE;
		else if ( !strcmp(argv[i], "-mon-console-nokey") ||
			  !strcmp(argv[i], "-mk")                )
			cfg.mon_emu |= (MONEMU_CONSOLE|MONEMU_CONSOLE_NOKEY);
		else if ( !strcmp(argv[i], "-mon-ignore-init") )
			cfg.mon_emu |= MONEMU_IGNORE_INIT;
		else if ( !strcmp(argv[i], "-sdx-tracks") )
			{
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%i", &cfg.tracks_sdxfdc[0]);
			}
		else if ( !strcmp(argv[i], "-sdx-mfloppy") )
			{
			if ( ++i == argc )
				usage();
			cfg.fn_sdxfdc[0] = argv[i];
			}
		else if ( !strcmp(argv[i], "-sdx-tracks2") )
			{
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%i", &cfg.tracks_sdxfdc[1]);
			}
		else if ( !strcmp(argv[i], "-sdx-mfloppy2") )
			{
			if ( ++i == argc )
				usage();
			cfg.fn_sdxfdc[1] = argv[i];
			}
		else if ( !strcmp(argv[i], "-sidisc-file") )
			{
			int drive;
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%i", &drive);
			if ( drive < 0 || drive >= N_SIDISC )
				fatal("bad Silicon Disc number");
			if ( ++i == argc )
				usage();
			cfg.sid_fn[drive] = argv[i];
			sid_set_file(drive, argv[i]);
			}
		else if ( !strcmp(argv[i], "-prn-file") )
			{
			if ( ++i == argc )
				usage();
			cfg.fn_print = argv[i];
			}
		else if ( !strcmp(argv[i], "-tape-dir") )
			{
			if ( ++i == argc )
				usage();
			if ( strlen(argv[i]) > MAX_TAPE_PREFIX )
				fatal("tape prefix is too long");
			cfg.tape_name_prefix = argv[i];
			}
		else if ( !strcmp(argv[i], "-tape-overwrite") )
			tape_overwrite = TRUE;
		else if ( !strcmp(argv[i], "-tape-disable") )
			cfg.tape_disable = TRUE;
		else if ( !strcmp(argv[i], "-cpm") )
			{
			cpm_init();
			mem_set_iobyte(0x80);
			addr = 0x100;
			}
		else if ( !strcmp(argv[i], "-cpm-drive-a") )
			{
			if ( ++i == argc )
				usage();
			cpm_set_drive_a(argv[i]);
			}
		else if ( !strcmp(argv[i], "-cpm-invert-case") )
			cpm_set_invert_case();
		else if ( !strcmp(argv[i], "-cpm-tail") )
			{
			if ( ++i == argc )
				usage();
			cpm_set_tail(argv[i]);
			}
		else if ( !strcmp(argv[i], "-cpm-open-hack") )
			cpm_allow_open_hack();
		else if ( !strcmp(argv[i], "-sdx") )
			setup_sdx();
		else if ( !strcmp(argv[i], "-speed") )
			{
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%ld", &clk_speed);
			if ( clk_speed < 1000000 || clk_speed > 1000000000 )
				fatal("clock speed must be between 1000000 (1MHz) and 1000000000 (1GHz)");
			}
		else if ( !strcmp(argv[i], "-fast") )
			moderate_speed = FALSE;
		else if ( !strcmp(argv[i], "-iperiod") )
			/* Undocumented feature.
			   I use this to investigate when I suspect a program
			   may be very timing sensitive. */
			{
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%d", &cfg.iperiod);
			if ( cfg.iperiod < 10 || cfg.iperiod > 1000000 )
				fatal("iperiod must be between 10 and 1000000");
			}
		else if ( !strcmp(argv[i], "-loadmtx") )
			loadmtx_hack = TRUE;
		else if ( !strcmp(argv[i], "-diag-z80-instructions-exclude") )
			{
			int addr_from, addr_to, addr;
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%i-%i", &addr_from, &addr_to);
			for ( addr = addr_from; addr <= addr_to; addr++ )
				no_trace[(addr&0xffff)>>3] |= (0x01<<(addr&7));
			}
		else if ( !strcmp(argv[i], "-diag-z80-instructions-include") )
			{
			int addr_from, addr_to, addr;
			if ( ++i == argc )
				usage();
			sscanf(argv[i], "%i-%i", &addr_from, &addr_to);
			for ( addr = addr_from; addr <= addr_to; addr++ )
				no_trace[(addr&0xffff)>>3] &= ~(0x01<<(addr&7));
			}
		else if ( !strcmp(argv[i], "-diag-ui-mem") )
			cfg.ui_opts |= UI_MEM;
		else if ( !strcmp(argv[i], "-diag-ui-vram") )
			cfg.ui_opts |= UI_VRAM;
		else if ( !strcmp(argv[i], "-diag-ui-dis") )
			cfg.ui_opts |= UI_DIS;
		else if ( !strncmp(argv[i], "-diag-", 6) )
			{
			int m;
			if ( (m = diag_method_of(argv[i]+6)) != 0 )
				diag_methods |= m;
			else if ( diag_flag_of(argv[i]+6) )
				;
			else
				usage();
			}
		else
			break;
		}

	/* Now try to support some convenient shorthands */
	/* printf ("argc = %d, i = %d\n", argc, i); */
	if ( i < argc )
		{
		const char *dot = strrchr(argv[i], '.');
		if ( dot == NULL )
			usage();
		if ( !strcmp(dot, ".com") || !strcmp(dot, ".COM") )
			{
			byte buf[0xfe00-0x0100];
			int len = read_file(argv[i], buf, sizeof(buf));
			char tail[128+1];
			cpm_init();
			mem_set_iobyte(0x80);
			mem_write_block(0x0100, len, buf);
			addr = 0x0100;
			++i;
			tail[0] = '\0';
			tail[1] = '\0';
			for ( ; i < argc; i++ )
				{
				if ( strlen(tail) + 1 + strlen(argv[i]) > 128 )
					fatal("command tail too long");
				strcat(tail, " ");
				strcat(tail, argv[i]);
				}
			cpm_set_tail( (tail[1]!='\0') ? tail : "" );
			}
		else if ( !strcmp(dot, ".run") || !strcmp(dot, ".RUN") )
			{
			int len;
			run_buf = emalloc(0x10000-0x4000);
			len = read_file(argv[i], run_buf, 0x10000-0x4000);
			if ( len < 4 )
				fatal("RUN file is way too short");
			run_hdr_base   = get_word(run_buf  );
			run_hdr_length = get_word(run_buf+2);
			if ( len < run_hdr_length )
				/* Per SDX User Manual, the length in the
				   header is doesn't include the header. */
				fatal("RUN file is too short");
			if ( run_hdr_base < 0x4000 ||
			     (unsigned long) run_hdr_base + (unsigned long) run_hdr_length >= 0x10000 )
				fatal("RUN file base and length fields aren't credible"); 
			mem_set_iobyte(0x00);
			mem_write_byte(0x3627, 0xed); /* was 0xd3 */
			mem_write_byte(0x3628, 0xfe); /* was 0x05 */
			addr = 0x0000;
			++i;
			fflush(stdout);
			}
		else if ( !strcmp(dot, ".mtx") || !strcmp(dot, ".MTX") )
			/* Save the filename for later LOAD "" */
			cfg.tape_fn = argv[i++];
		}

	if ( i != argc )
		usage();

#ifndef NO_TH
	if ( (diag_methods & DIAGM_CONSOLE) != 0 &&
	     (cfg.mon_emu & MONEMU_TH) != 0 )
		fatal("cannot emulate the monitor using TH and send diagnostics to the console at the same time");
	if ( (cfg.mon_emu & MONEMU_TH     ) != 0 &&
	     (cfg.mon_emu & MONEMU_CONSOLE) != 0 ) 
		fatal("cannot emulate the monitor using TH and the console at the same time");
#endif

	{
	time_t t_start = time(NULL);
	char buf[24+1];
	strcpy(buf, ctime(&t_start));
	buf[24] = '\0';
	diag_message(DIAG_ALWAYS, "=== memu started at %s ===", buf);
	}

	ctc_init();
	vid_init(cfg.vid_emu, cfg.vid_width_scale, cfg.vid_height_scale);
	kbd_init(cfg.kbd_emu);
	snd_init(cfg.snd_emu, cfg.latency);
	mon_init(cfg.mon_emu, cfg.mon_width_scale, cfg.mon_height_scale);

	if ( cfg.fn_sdxfdc[0] != NULL )
		{
		sdxfdc_drvcfg(0, sdxcfg(cfg.tracks_sdxfdc[0]));
		sdxfdc_init(0, cfg.fn_sdxfdc[0]);
		}
	if ( cfg.fn_sdxfdc[1] != NULL )
		{
		sdxfdc_drvcfg(1, sdxcfg(cfg.tracks_sdxfdc[1]));
		sdxfdc_init(1, cfg.fn_sdxfdc[1]);
		}

	sid_init();

	if ( cfg.fn_print != NULL )
		print_init(cfg.fn_print);

	ui_init(cfg.ui_opts);

	if ( !cfg.tape_disable )
		{
		byte b = mem_get_iobyte();
		mem_set_iobyte(0x00);
		mem_write_byte(0x0aae, 0xed);
		mem_write_byte(0x0aaf, 0xfe);
		mem_write_byte(0x0ab0, 0xc9);
		mem_set_iobyte(b);
		}

	ResetZ80(&z80);
	z80.PC.W = (word) addr;
	z80.Trace = 1;
	z80.IPeriod = cfg.iperiod;
	RunZ80(&z80);

	return 0;
	}
/*...e*/
