/* rpi.c - Provides routines specific to memu-pi */

#include "types.h"
#include "diag.h"
#include "config.h"
#include "common.h"

#include <unistd.h>
/* #include <linux/reboot.h> */
#include <sys/reboot.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <termios.h>

int memu_entry (int argc, char *argv[]);

/* A replacement for the normal exit () system call, which halts the system */

void rpi_exit (int iStatus)
   {
   int iExit;
   /* Sync the file system to ensure all pending writes have completed */

   printf ("Memu-Pi shutting down.\n");
   sync ();
   sleep (2);
   sync ();
   sleep (2);
   printf ("Filesystem sinced.\n");

   /* Give time to read error message */

   if ( iStatus != 0 )  sleep (10);

   /* Halt the system */

   // printf ("Shutting down\n");
/*   iExit = reboot (LINUX_REBOOT_MAGIC1, LINUX_REBOOT_MAGIC2, LINUX_REBOOT_CMD_POWER_OFF, NULL); */
   iExit = reboot (RB_POWER_OFF);
   perror ("reboot: ");

   /* Failed to power down - wait forever */

   while (TRUE) sleep (1000000);
   }

static void uk_kbd_map (void)
	{
	static int ttyfd  =  0;
	struct kbentry kbe;
	struct
		{
		int	code;
		int shift;
		int key;
		} uk_map[] = {
		{ 0x03, K_SHIFTTAB, '"' },
			{ 0x28,  K_NORMTAB, '\'' },
			{ 0x28, K_SHIFTTAB, '@' },
			{ 0x2b, K_NORMTAB, '#' },
			{ 0x2b, K_SHIFTTAB, '~' },
			{ 0x56, K_NORMTAB, '\\' },
			{ 0x56, K_SHIFTTAB, '|' } };
	int i;

	for ( i = 0; i < sizeof (uk_map) / sizeof (uk_map[0]); ++i )
		{
		kbe.kb_table   =  uk_map[i].shift;
		kbe.kb_index   =  uk_map[i].code;
		kbe.kb_value   =  uk_map[i].key;
		if ( ioctl (ttyfd, KDSKBENT, &kbe) < 0 )
			diag_message (DIAG_HW_KBD, "Failed to map %s key 0x%02x to \'%c\'",
			( uk_map[i].shift == K_SHIFTTAB ? "Shifted" : "Normal" ), uk_map[i].code, uk_map[i].key);
		}
	}

int main (int argc, char *argv[])
	{
	// printf ("argv[0] = %s\n", argv[0]);
	char *psCfg = cfg_find_file (argv[0]);
    // printf ("psCfg = %s\n", psCfg);
	if ( psCfg == NULL )
		{
		// printf ("Trying default configuration file.\n");
		psCfg = estrdup ("/memu/memu-pi.cfg");
		if ( ! cfg_test_file (psCfg) ) fatal ("No configuration file found");
		}
	if ( ! read_config (psCfg, &argc, &argv) ) fatal ("Error reading configuration file");
	// printf ("Read configuration file.\n");
	free (psCfg);
	sleep (2);
	// printf ("Calling memu_entry.\n");
	uk_kbd_map ();
	memu_entry (argc, argv);
	}
